#include <string.h>
#include <stralloc.h>
#include <afblib/inbuf.h>
#include <afblib/outbuf.h>
#include "lockmanager.h"
#include "mxprequest.h"
#include "mxpresponse.h"
#include "mxpsession.h"

#define EQUAL(sa,str) (strncmp((sa.s),(str),(sa.len)) == 0)

void mxp_session(int fd, int argc, char** argv) {
   if (argc != 1) return;
   char* lockdir = argv[0];

   inbuf ibuf = {fd};
   outbuf obuf = {fd};
   lockset locks = {0};

   /* send greeting */
   mxp_response greeting = {MXP_SUCCESS};
   if (!write_mxp_response(&obuf, &greeting)) return;
   if (!outbuf_flush(&obuf)) return;

   /* receive identification */
   mxp_request id = {{0}};
   if (!read_mxp_request(&ibuf, &id)) return;
   if (!EQUAL(id.keyword, "id")) return;
   stralloc myname = {0};
   stralloc_copy(&myname, &id.parameter);
   stralloc_0(&myname);
   int ok = lm_init(&locks, lockdir, myname.s);

   /* send response to identification */
   mxp_response response = {MXP_SUCCESS};
   stralloc_copys(&response.message, "welcome");
   if (!ok) response.status = MXP_FAILURE;
   if (!write_mxp_response(&obuf, &response)) return;
   if (!outbuf_flush(&obuf)) return;
   if (!ok) return;

   /* process regular requests */
   mxp_request request = {{0}};
   while (read_mxp_request(&ibuf, &request)) {
      stralloc lockname = {0};
      stralloc_copy(&lockname, &request.parameter);
      stralloc_0(&lockname);

      if (EQUAL(request.keyword, "stat")) {
         mxp_response info = {MXP_CONTINUATION};
         if (lm_stat(&locks, lockname.s, &info.message)) {
            response.status = MXP_SUCCESS;
            if (info.message.len == 0) {
               stralloc_copys(&response.message, "free");
            } else {
               if (!write_mxp_response(&obuf, &info)) break;
               stralloc_copys(&response.message, "held");
            }
         } else {
            response.status = MXP_FAILURE;
            stralloc_copys(&response.message,
               "unable to check lock status");
         }
         free_mxp_response(&info);
      } else if (EQUAL(request.keyword, "lock")) {
         if (lm_nonblocking_lock(&locks, lockname.s)) {
            response.status = MXP_SUCCESS;
            stralloc_copys(&response.message, "locked");
         } else {
            mxp_response notification = {MXP_CONTINUATION};
            stralloc_copys(&notification.message, "waiting");
            if (!write_mxp_response(&obuf, &notification)) break;
            if (!outbuf_flush(&obuf)) break;
            if (lm_lock(&locks, lockname.s)) {
               response.status = MXP_SUCCESS;
               stralloc_copys(&response.message, "locked");
            } else {
               response.status = MXP_FAILURE;
               stralloc_copys(&response.message, "");
            }
         }
      } else if (EQUAL(request.keyword, "release")) {
         stralloc_copys(&response.message, "");
         if (lm_release(&locks, lockname.s)) {
            response.status = MXP_SUCCESS;
         } else {
            response.status = MXP_FAILURE;
         }
      } else {
         response.status = MXP_FAILURE;
         stralloc_copys(&response.message, "unknown command");
      }
      if (!write_mxp_response(&obuf, &response)) break;
      if (!outbuf_flush(&obuf)) break;
   }
   /* release all locks */
   lm_free(&locks);
   /* free allocated memory */
   free_mxp_response(&response);
   stralloc_free(&myname);
}
